/*
 * Christian Benjamin Ries, November 2010
 * Christian_Benjamin.Ries@fh-bielefeld.de
 * http://www.christianbenjaminries.de
 * University of Applied Sciences Bielefeld, Germany
 * License: GNU GPL
 */

// C++
#include <iostream>
#include <string>
#include <vector>

// C
#include <stdio.h>
#include <errno.h>

// Third-party
#include <zip.h>
#include <Magick++.h>

// BOINC
#include <boinc_api.h>
#include <filesys.h>
#include <parse.h>
#include <mfile.h>
#include <miofile.h>
#include <error_numbers.h>

#define CHECKPOINT_FILE 		"lmboinc.chk"
#define CONFIGURATION_FILE 	"lmboinc.xml"
#define INPUT_FILE 					"archive_in.zip"
#define OUTPUT_FILE 				"archive_out.zip"

typedef std::vector<std::string> FileList;

/**
 * @brief Describes configuration values for runtime.
 * @method init Initializes embedded variables with default values.
 * @method parse Parses a XML-Files and filters values and sets the variables with these values.
<b>Example:</b>
<pre>
<lmboinc>
  <mode>edge</mode>
  <copyright>Christian Benjamin Ries</copyright>
</lmboinc>
</pre>
 */
typedef struct LMBoincXML {
	std::string mode; /*!< Describes manipulation mode, e.q. edge, oil, negate, shape, normalize. */
	std::string copyright; /*!< Value of copyright string at top of images. */
	std::string size; /*!< e.q. 320x240 @see http://www.imagemagick.org/Magick++/Image.html */
	std::string type; /*!< e.q. RGBA @see http://www.imagemagick.org/Magick++/Image.html */
	
	inline void init() {
		this->mode = "edge";
		this->copyright = "Christian Benjamin Ries";
		this->size = "320x240";
		this->type = "RGBA";
	}
	
	/**
	 * @param lmBoincXMLFiles Pointer to open xml file.
	 * @return -1 if something went wrong, otherwise 0.
	 */
	int parse(FILE *lmBoincXMLFile) {
		init();

		MIOFILE xmlMIOFile;
		xmlMIOFile.init_file(lmBoincXMLFile);
		XML_PARSER p(&xmlMIOFile);

		if(!p.parse_start("lmboinc")) {
			fprintf(stderr, "lmboinc.xml is not formatted correctly\n");
			return ERR_XML_PARSE;
		}

		bool is_tag = false;
		char tag[128] = {'\0'};
		while (!p.get(tag, sizeof(tag), is_tag)) {
			if (!is_tag) {
				fprintf(stderr, "unexpected text in lmboinc.xml: %s\n", tag);
				continue;
			}
			if (!strcmp(tag, "/lmboinc")) return 0;
			if (p.parse_string(tag, "mode", this->mode)) continue;
			if (p.parse_string(tag, "copyright", this->copyright)) continue;
			if (p.parse_string(tag, "size", this->size)) continue;
			if (p.parse_string(tag, "type", this->type)) continue;
		}
		return 0;
	}
} LMBoincXML;


/**
 * @param configuration
 * @return FALSE if something went wrong, otherwise TRUE.
 */
bool initLMBoincXML(LMBoincXML &configuration) {
	std::string xmlFilePath;
  boinc_resolve_filename_s(CONFIGURATION_FILE, xmlFilePath);

	FILE *xmlFile = boinc_fopen(xmlFilePath.c_str(), "r");
	if(xmlFile == NULL) {
		return false;
	}
	return (configuration.parse(xmlFile)?false:true);
}


/**
 * Removes files from hard disk.
 * @param files Names of files to remove.
 */
void removeOldFiles(FileList *files) {
	if(files != NULL) {
		FileList::iterator it = files->begin();
		for( ; it != files->end(); it++ ) {
			boinc_delete_file(it->c_str());
		}
		files->clear();
	}
}


/**
 * @param files
 * @param countFilesDone
 */
void doCheckpoint(FileList *files, int countFilesDone) {
	std::string checkpointPath;
	FILE *checkpointFile = NULL;
	std::string outputArchivePath;
	struct zip *zipFileOut = NULL;
	char zipError[512] = {'\0'};
	int returnValue = 0;
	
	// Den Pfad im Dateisystem zur Archiv-Datei ermitteln und oeffnen.
	boinc_resolve_filename_s(OUTPUT_FILE, outputArchivePath);
	zipFileOut = zip_open(outputArchivePath.c_str(), ZIP_CREATE, NULL);
	if (zipFileOut == NULL) {
		zip_error_to_str(zipError, 512, 0, errno);
		std::cerr << "zip failed: " << zipError << std::endl;
		boinc_checkpoint_completed();
		boinc_finish(-1);
	}

	// Die Dateien seit dem Start der Anwendung oder seit dem 
	// letzten Erstellen einer checkpoint-Datei zum Archiv
	// hinzufuegen.
	FileList::iterator it = files->begin();
	for( ; it != files->end(); it++ ) 
	{
		struct zip_source *s = NULL;
		if( (s=zip_source_file(zipFileOut, it->c_str(), 0, 0)) == NULL
				|| zip_add(zipFileOut, it->c_str(), s) < 0) 
		{ 
						zip_source_free(s); 
						std::cout << "zip failed: error adding file " << zip_strerror(zipFileOut) << std::endl; 
		}
	}
	zip_close(zipFileOut);

	// Den Pfad im Dateisystem zur Checkpoint-Datei ermitteln und
	// die aktuellen Anzahl der bisher bearbeiteten Dateien fuer 
	// eine spaetere Verwendung speichern.
	boinc_resolve_filename_s(CHECKPOINT_FILE, checkpointPath);
	checkpointFile = boinc_fopen(checkpointPath.c_str(), "w+");
	if (checkpointFile != NULL) {
		returnValue = fprintf(checkpointFile, "%d\n", countFilesDone);
	}
	fclose(checkpointFile);

	// Die hinzugefuegten Dateien von der Festplatte loeschen.
	removeOldFiles(files);
}


/**
 * @param argc
 * @param argv
 * @return
 */
int main(int argc, char **argv) {
		// Magick++
		Magick::InitializeMagick(*argv);
		
		// Das BOINC framework fuer die Ausfuehrung
		// dieser Applikation initialisieren.
		int returnValue = boinc_init();
    if (returnValue) {
				char buf[128] = {'\0'};
				std::cerr << boinc_msg_prefix(buf) << " boinc_init returned " << returnValue << std::endl;
				exit(returnValue);
    }

		LMBoincXML lmBoincConfiguration;
		if(!initLMBoincXML(lmBoincConfiguration)) {
			// no message
		}

		// Dateinamen durch den BOINC client ausloesen und
		// das Archiv zum Lesen oeffnen.
		std::string inputArchivePath;
		// Die passenden Modifikationen fuer Windows und ZIP-Support 
 		// muessen noch vorgenommen werden.
		struct zip *zipFileIn = NULL;
		char zipError[512] = {'\0'};
		
    boinc_resolve_filename_s(INPUT_FILE, inputArchivePath);
		zipFileIn = zip_open(inputArchivePath.c_str(), ZIP_CHECKCONS, NULL);
		if (zipFileIn == NULL) {
				zip_error_to_str(zipError, 512, 0, errno);
				std::cerr << "zip failed: " << zipError << std::endl;
        boinc_finish(-1);
    }

		// Anzahl der im Archiv enthaltenen Dateien ermitteln,
		// wichtig fuer die Berechnung der aktuellen Bearbeitung in Prozent.
		int countFilesIn = zip_get_num_files(zipFileIn);
		int countFilesDone = 0;

		// Eventuell vorhandene checkpoint-Datei oeffnen und
		// den enthaltenen Zahlenwert auslesen. Dieser Zahlenwert
		// gibt Auskunft ueber die bisher bearbeiteten Bilddateien aus
		// dem Archiv.
		std::string checkpointPath;
		FILE *checkpointFile = NULL;
		
		boinc_resolve_filename_s(CHECKPOINT_FILE, checkpointPath);
		checkpointFile = boinc_fopen(checkpointPath.c_str(), "r");
		if (checkpointFile != NULL) {
			returnValue = fscanf(checkpointFile, "%d", &countFilesDone);
			if (returnValue <= 0) {
				std::cerr << "checkpoint failed: Could not read in current check point." << std::endl;
				countFilesDone = 0;
			}
			fclose(checkpointFile);
		}
		
		if (countFilesDone == countFilesIn) {
			boinc_fraction_done(1);
			boinc_finish(0);
		}
		
		/*
		 * Hier faengt die Magie 'Magick++' an :-)
		 */
		FileList fileList;

		// Alle Dateien aus dem Archiv lesen.
		while (countFilesDone < countFilesIn) {
			
			// Eine Datei aus dem Archiv selektieren.
			struct zip_file *zipReader = zip_fopen_index(zipFileIn, countFilesDone, ZIP_FL_UNCHANGED);
			if (zipReader == NULL) {
				zip_error_to_str(zipError, 512, 0, errno);
				std::cerr << "zip failed: " << zipError << std::endl;
				boinc_finish(-1);
			}

			// Informationen ueber eine Datei im zip-Archiv ermitteln.
			struct zip_stat zipFileInformation;
			zip_stat_index(zipFileIn, countFilesDone, ZIP_FL_UNCHANGED, &zipFileInformation);
			
			// Datei aus dem Archiv auslesen.
			char *imageBuffer = new char[zipFileInformation.size];
			zip_fread(zipReader, imageBuffer, zipFileInformation.size);
						
			// Die Datei in ein Magick-Objekt umwandeln
			// und die Bilddaten manipulieren.
			Magick::Blob blob((void*)imageBuffer, zipFileInformation.size);
			Magick::Image img;
			img.size(lmBoincConfiguration.size);
			img.magick(lmBoincConfiguration.type);
			img.read(blob);
			if 				(lmBoincConfiguration.mode == "edge") {
				img.edge();
			} else if (lmBoincConfiguration.mode == "oil") {
				img.oilPaint();
			} else if (lmBoincConfiguration.mode == "shade") {
				img.shade();
			} else if (lmBoincConfiguration.mode == "negate") {
				img.negate();
			} else if (lmBoincConfiguration.mode == "normalize") {
				img.normalize();
			} else {
				;
			}
			// Text zum Bild hinzufuegen.
			img.fillColor("white"); // Fill color
			img.fontPointsize(30);
			img.draw( Magick::DrawableText(10, 35, lmBoincConfiguration.copyright) );
			
			// Datei auf die Festplatte schreiben.
			img.write(zipFileInformation.name);
			
			// Namen merken und bei der Erstellung des
			// Checkpunktes oder beim Ende der Programmausfuehrung
			// die aufgefuehrten Dateien von der Festplatte loeschen.
			fileList.push_back(zipFileInformation.name);

			if (imageBuffer != NULL) {
				delete imageBuffer; imageBuffer = NULL;
			}

			zip_fclose(zipReader);

			countFilesDone++;

			// Berechnung der bisher geloesten Aufgaben und
			// Uebertragung des Zahlenwertes an den BOINC client.
			double done = countFilesDone/(double)countFilesIn;		
			boinc_fraction_done(done);
			//boinc_report_app_status(0.0, 0.0, done);	
	
			// Den aktuellen Zwischenstand in einer checkpoint-Datei merken.
			if (boinc_time_to_checkpoint()) {
		    doCheckpoint(&fileList, countFilesDone);				
				boinc_checkpoint_completed();
			}
		}

		doCheckpoint(&fileList, countFilesDone);
		/*
		 * Och, die Zaubershow ist schon vorbei?
		 */
		
    boinc_fraction_done(1);
    boinc_finish(0);
}


#ifdef _WIN32
int WINAPI WinMain(HINSTANCE hInst, HINSTANCE hPrevInst, LPSTR Args, int WinMode) {
    LPSTR command_line;
    char* argv[100];
    int argc;

    command_line = GetCommandLine();
    argc = parse_command_line( command_line, argv );
    return main(argc, argv);
}
#endif
